# fsl_sub plugin for:
#  * Sun Grid Engine
#  * Son of Grid Engine
#  * Open Grid Scheduler
#  * Univa Grid Engine
import datetime
import logging
import os
import shlex
import subprocess as sp
import tempfile
import xml.etree.ElementTree as ET
from collections import defaultdict
from functools import lru_cache
from shutil import which

from fsl_sub.exceptions import (
    BadSubmission,
    BadConfiguration,
    GridOutputError,
    UnknownJobId,
)
from fsl_sub.config import (
    method_config,
    coprocessor_config,
    read_config,
)
import fsl_sub.consts
from fsl_sub.utils import (
    split_ram_by_slots,
    parse_array_specifier,
)


def plugin_version():
    return '1.2.7'


def qtest():
    '''Command that confirms method is available'''
    return qconf_cmd()


def qconf_cmd():
    '''Command that queries queue configuration'''
    qconf = which('qconf')
    if qconf is None:
        raise BadSubmission("Cannot find Grid Engine software")
    return qconf


@lru_cache()
def qstat_cmd():
    '''Command that queries queue state'''
    qstat = which('qstat')
    if qstat is None:
        raise BadSubmission("Cannot find Grid Engine software")
    return qstat


@lru_cache()
def qacct_cmd():
    '''Command that queries completed job stats'''
    qacct = which('qacct')
    if qacct is None:
        raise BadSubmission("Cannot find Grid Engine software")
    return qacct


def qsub_cmd():
    '''Command that submits a job'''
    qsub = which('qsub')
    if qsub is None:
        raise BadSubmission("Cannot find Grid Engine software")
    return qsub


def queue_exists(qname, qtest=None):
    '''Does qname exist'''
    if qtest is None:
        qtest = qconf_cmd()
    try:
        sp.run(
            [qtest, '-sq', qname],
            stdout=sp.DEVNULL,
            stderr=sp.DEVNULL,
            check=True, universal_newlines=True)
    except FileNotFoundError:
        raise BadSubmission(
            "Grid Engine software may not be correctly installed")
    except sp.CalledProcessError:
        return False
    return True


def check_pe(pe_name, queue, qconf=None, qstat=None):
    if qconf is None:
        qconf = qconf_cmd()
    if qstat is None:
        qstat = qstat_cmd()
    # Check for configured PE of pe_name
    cmd = sp.run(
        [qconf, '-sp', pe_name],
        stdout=sp.DEVNULL,
        stderr=sp.DEVNULL)
    if cmd.returncode != 0:
        raise BadSubmission(pe_name + " is not a valid PE")

    # Check for availability of PE
    cmd = sp.run(
        [qstat, '-g', 'c', '-pe', pe_name, '-xml'],
        universal_newlines=True,
        stdout=sp.PIPE, stderr=sp.PIPE)
    if (cmd.returncode != 0 or (
            cmd.stderr is not None and
            "error: no such parallel environment" in cmd.stderr)):
        raise BadSubmission(
            "No instances of {} configured".format(pe_name))

    # Check that PE is available on requested queue
    queue_defs = ET.fromstring(cmd.stdout)
    if queue not in [b.text for b in queue_defs.iter('name')]:
        raise BadSubmission(
            "PE {} is not configured on {}".format(pe_name, queue)
        )


def _valid_resources(res_defs):
    validity = [True if a else False for a in res_defs if '=' in a]
    return all(validity)


def _resource_dict(res):
    '''Returns a dictionary of key, value from a list of strings of form
    key=value'''

    return dict(a.split('=') for a in res)


def _filter_on_resource_dict(res_list, res_defs):
    '''Return list of resources from res_list that aren't present in the
    resource definitions in res_defs dict'''
    if res_defs is None:
        return res_list
    result = [r for r in res_list if r not in res_defs]
    return result


def submit(
        command,
        job_name,
        queue,
        threads=1,
        array_task=False,
        jobhold=None,
        array_hold=None,
        array_limit=None,
        array_specifier=None,
        parallel_env=None,
        jobram=None,
        jobtime=None,
        resources=None,
        ramsplit=False,
        priority=None,
        mail_on=None,
        mailto=None,
        logdir=None,
        coprocessor=None,
        coprocessor_toolkit=None,
        coprocessor_class=None,
        coprocessor_class_strict=False,
        coprocessor_multi=1,
        usescript=False,
        architecture=None,
        requeueable=True,
        project=None
        ):
    '''Submits the job to a Grid Engine cluster
    Requires:

    command - string or list containing command to run
                or the file name of the array task file.
                If array_specifier is given then this must be
                a string/list containing the command to run.
    job_name - Symbolic name for task
    queue - Queue to submit to

    Optional:
    array_task - is the command is an array task (defaults to False)
    jobhold - id(s) of jobs to hold for (string or list)
    array_hold - complex hold string
    array_limit - limit concurrently scheduled array
            tasks to specified number
    array_specifier - n[-m[:s]] n subtasks or starts at n, ends at m with
            a step of s
    parallelenv - parallel environment name
    jobram - RAM required by job (total of all threads)
    jobtime - time (in minutes for task)
    requeueable - may a job be requeued if a node fails
    resources - list of resource request strings
    ramsplit - break tasks into multiple slots to meet RAM constraints
    priority - job priority (0-1023)
    mail_on - mail user on 'a'bort or reschedule, 'b'egin, 'e'nd,
            's'uspended, 'n'o mail
    mailto - email address to receive job info
    logdir - directory to put log files in
    coprocessor - name of coprocessor required
    coprocessor_toolkit - coprocessor toolkit version
    coprocessor_class - class of coprocessor required
    coprocessor_class_strict - whether to choose only this class
            or all more capable
    coprocessor_multi - how many coprocessors you need (or
            complex description) (string)
    usescript - queue config is defined in script
    project - which project to associate this job with
    '''

    logger = logging.getLogger('fsl_sub.plugins')
    if command is None:
        raise BadSubmission(
            "Must provide command line or array task file name")

    mconf = defaultdict(lambda: False, method_config('sge'))
    qsub = qsub_cmd()
    command_args = [qsub, ]

    if isinstance(resources, str):
        resources = [resources, ]

    os.environ['FSLSUB_ARRAYTASKID_VAR'] = 'SGE_TASK_ID'
    os.environ['FSLSUB_ARRAYSTARTID_VAR'] = 'SGE_TASK_FIRST'
    os.environ['FSLSUB_ARRAYENDID_VAR'] = 'SGE_TASK_LAST'
    os.environ['FSLSUB_ARRAYSTEPSIZE_VAR'] = 'SGE_TASK_STEPSIZE'
    os.environ['FSLSUB_ARRAYCOUNT_VAR'] = ''

    if type(queue) == str:
        if ',' in queue:
            queue = queue.split(',')
        else:
            queue = [queue, ]

    if usescript:
        if not isinstance(command, str):
            raise BadSubmission(
                "Command should be a grid submission script (no arguments)")
    else:
        # Check Parallel Environment is available
        if parallel_env:
            for q in queue:
                check_pe(parallel_env, q)
            command_args.extend(
                ['-pe', parallel_env, str(threads), '-w', 'e'])
            command_args.extend(
                ['-R', 'y', ])
        if mconf['copy_environment']:
            command_args.append('-V')

        binding = mconf['affinity_type']

        if coprocessor is not None:
            # Setup the coprocessor
            cpconf = coprocessor_config(coprocessor)
            if cpconf['no_binding']:
                binding = None
            if cpconf['classes']:
                available_classes = cpconf['class_types']
                if coprocessor_class is None:
                    coprocessor_class = cpconf['default_class']
                if (coprocessor_class_strict or
                        not cpconf['include_more_capable']):
                    try:
                        copro_class = available_classes[
                                            coprocessor_class][
                                                'resource']
                    except KeyError:
                        raise BadSubmission("Unrecognised coprocessor class")
                else:
                    copro_capability = available_classes[
                                            coprocessor_class][
                                                'capability'
                                            ]
                    base_list = [
                        a for a in cpconf['class_types'].keys() if
                        cpconf['class_types'][a]['capability'] >=
                        copro_capability]
                    copro_class = '|'.join(
                        [
                            cpconf['class_types'][a]['resource'] for a in
                            sorted(
                                    base_list,
                                    key=lambda x:
                                    cpconf['class_types'][x]['capability'])
                        ]
                    )

                command_args.extend(
                    ['-l',
                     '='.join(
                          (cpconf['class_resource'], copro_class))]
                         )
            command_args.extend(
                ['-l',
                 '='.join(
                      (cpconf['resource'], str(coprocessor_multi)))]
                    )

        if binding is not None:
            if mconf['affinity_control'] == 'threads':
                affinity_spec = ':'.join(
                    (mconf['affinity_type'], str(threads)))
            elif mconf['affinity_control'] == 'slots':
                affinity_spec = ':'.join(
                    (mconf['affinity_type'], 'slots'))
            else:
                raise BadConfiguration(
                    ("Unrecognised affinity_control setting " +
                     mconf['affinity_control']))
            command_args.extend(['-binding', affinity_spec])

        if (mconf['job_priorities'] and
                priority is not None):
            if 'min_priority' in mconf:
                priority = max(mconf['min_priority'], priority)
                priority = min(mconf['max_priority'], priority)
            command_args.extend(['-p', str(priority), ])

        if resources and _valid_resources(resources):
            command_args.extend(
                ['-l', ','.join(resources), ])
            resources = _resource_dict(resources)

        if logdir is not None:
            command_args.extend(
                ['-o', logdir, '-e', logdir]
            )

        hold_state = '-hold_jid'
        # But have I not specified that array_task is not set for array
        # aware tasks
        if array_task and array_hold:
            if mconf['array_holds']:
                jobhold = array_hold
                hold_state = '-hold_jid_ad'
            else:
                jobhold = array_hold

        if jobhold:
            if isinstance(jobhold, (list, tuple)):
                parents = ','.join([str(a) for a in jobhold])
            elif isinstance(jobhold, str):
                parents = jobhold
            elif isinstance(jobhold, int):
                parents = str(jobhold)
            else:
                raise BadSubmission(
                    "jobhold is of unsupported type " + str(type(jobhold)))
            command_args.extend(
                [hold_state, parents]
                )

        if array_task is not None:
            if mconf['array_limit'] and array_limit:
                command_args.extend(
                    ['-tc', str(array_limit), ]
                )

        if jobram:
            ram_units = read_config()['ram_units']
            if ramsplit:
                jobram = split_ram_by_slots(jobram, threads)
            if mconf['notify_ram_usage']:
                # If a ram_resource has been specified for the job don't change
                ram_resources = sorted(
                    _filter_on_resource_dict(
                        mconf['ram_resources'],
                        resources))
                if ram_resources:
                    command_args.extend(
                        ['-l', ','.join(
                            ['{0}={1}{2}'.format(
                                a, jobram, ram_units) for
                                a in ram_resources])]
                    )

        if mconf['mail_support']:
            if mailto:
                command_args.extend(['-M', mailto, ])
                if not mail_on:
                    mail_on = mconf['mail_mode']
                if mail_on not in mconf['mail_modes']:
                    raise BadSubmission("Unrecognised mail mode")
                command_args.extend(
                    [
                        '-m',
                        ','.join(mconf['mail_modes'][mail_on])
                        ])

        command_args.extend(['-N', job_name, ])
        command_args.extend(
            ['-cwd', '-q', ','.join(queue), ])

        if requeueable:
            command_args.extend(
                ['-r', 'y', ]
            )
        if project is not None:
            command_args.extend(
                ['-P', project, ]
            )
        if array_task:
            # Submit array task file
            if array_specifier:
                (
                    array_start,
                    array_end,
                    array_stride
                    ) = parse_array_specifier(array_specifier)
                if not array_start:
                    raise BadSubmission("array_specifier doesn't make sense")
                array_spec = "{0}". format(array_start)
                if array_end:
                    array_spec += "-{0}".format(array_end)
                if array_stride:
                    array_spec += ":{0}".format(array_stride)
                command_args.extend(['-t', array_spec])
            else:
                with open(command, 'r') as cmd_f:
                    array_slots = len(cmd_f.readlines())
                command_args.extend(
                    ['-t', "1-{0}".format(
                        array_slots)])
        if not array_task or array_specifier:
            # Submit single script/binary
            command_args.extend(
                ['-shell', 'n',
                 '-b', 'y', ])
            if isinstance(command, str):
                command = shlex.split(command)
            command_args.extend(command)

    logger.info("sge_args: " + " ".join(
        [str(a) for a in command_args if a != qsub]))

    if array_task and not array_specifier:
        logger.info("executing array task")
        scriptcontents = '''#!/bin/bash

#$ -S /bin/bash

the_command=$(sed -n -e "${{SGE_TASK_ID}}p" {0})

exec /bin/bash -c "$the_command"
'''.format(command)
        logger.debug(scriptcontents)
        with tempfile.NamedTemporaryFile(
                mode='wt',
                delete=False) as script:
            script.write(scriptcontents)
        logger.debug(script.name)
        command_args.append(script.name)
    else:
        if usescript:
            logger.info("executing cluster script")
        else:
            if array_specifier:
                logger.info("excuting array task {0}-{1}:{2}".format(
                    array_start,
                    array_end,
                    array_stride
                ))
            else:
                logger.info("executing single task")
            logger.info(" ".join([str(a) for a in command_args]))
    logger.debug(type(command_args))
    logger.debug(command_args)
    result = sp.run(
        command_args, universal_newlines=True,
        stdout=sp.PIPE, stderr=sp.PIPE)
    if array_task and not array_specifier:
        os.remove(script.name)
    if result.returncode != 0:
        raise BadSubmission(result.stderr)
    job_words = result.stdout.split(' ')
    try:
        job_id = int(job_words[2].split('.')[0])
    except ValueError:
        raise GridOutputError("Grid output was " + result.stdout)
    return job_id


def example_conf():
    '''Returns a string containing the example configuration for this
    cluster plugin.'''

    here = os.path.realpath(os.path.dirname(__file__))
    with open(os.path.join(here, 'fsl_sub_sge.yml')) as e_conf_f:
        e_conf = e_conf_f.read()

    return e_conf


def job_status(job_id, sub_job_id=None):
    '''Return details for the job with given ID.

    details holds a dict with following info:
        id
        name
        script (if available)
        arguments (if available)
        # sub_state: fsl_sub.consts.NORMAL|RESTARTED|SUSPENDED
        submission_time
        tasks (dict keyed on sub-task ID):
            status:
                fsl_sub.consts.QUEUED
                fsl_sub.consts.RUNNING
                fsl_sub.consts.FINISHED
                fsl_sub.consts.FAILEDNQUEUED
                fsl_sub.consts.HELD
            start_time
            end_time
            sub_time
            utime
            stime
            exit_status
            error_message
            maxmemory (in Mbytes)
        parents (if available)
        children (if available)
        job_directory (if available)

        '''

    # Look for running jobs
    if isinstance(job_id, str):
        if '.' in job_id:
            if sub_job_id is None:
                job_id, sub_job_id = job_id.split('.')
                sub_job_id = int(sub_job_id)
            else:
                job_id, _ = job_id.split('.')
        job_id = int(job_id)
    if isinstance(sub_job_id, str):
        sub_job_id = int(sub_job_id)

    try:
        job_details = _qstat_job(job_id, sub_job_id=sub_job_id)
        if job_details:
            return job_details
        else:
            job_details = _qacct_job(job_id, sub_job_id=sub_job_id)

    except UnknownJobId:
        raise
    except Exception as e:
        raise GridOutputError from e

    return job_details


def _qacct_job(job_id, sub_job_id=None):
    job_details = {}
    task_id = 1
    sub_task = False
    if sub_job_id is not None:
        sub_task = True
        task_id = sub_job_id

    try:
        job_info = _qacct_job_lines(job_id)
    except UnknownJobId:
        return None
    raw_task_blocks = {}
    block_id = -1
    for line in job_info:
        if line == '':
            continue
        if line.startswith('============'):
            # new task block
            block_id += 1
            raw_task_blocks[block_id] = {}
            continue
        line = ' '.join(line.strip().split())
        (key, value) = line.split(' ', maxsplit=1)
        raw_task_blocks[block_id][key] = value

    tasks = {}
    for tb in raw_task_blocks.values():
        tasks[tb['taskid']] = tb

    sample_task = tasks[list(tasks.keys())[0]]
    job_details['id'] = int(sample_task['jobnumber'])
    job_details[
        'submission_time'] = _qacct_timestamp(sample_task['qsub_time'])
    job_details['name'] = sample_task['jobname']
    try:
        job_details['job_directory'] = sample_task['cwd']
    except KeyError:
        job_details['job_directory'] = None

    job_details['parents'] = None
    job_details['children'] = None
    # On Univa this is the complete qsub command line
    try:
        job_details['script'] = sample_task['submit_cmd']
    except KeyError:
        job_details['script'] = None
    job_details['args'] = None

    if sub_task:
        sub_tasks = [t for t in tasks.values() if t['taskid'] == task_id]
        if not sub_tasks:
            raise UnknownJobId(task_id)
    else:
        sub_tasks = list(tasks.values())
    job_details['tasks'] = {}
    for details in sub_tasks:
        try:
            taskid = int(details['taskid'])
        except ValueError:
            taskid = 1
        job_details['tasks'][taskid] = {}
        td = job_details['tasks'][taskid]
        td['status'] = fsl_sub.consts.FINISHED
        td['exit_status'] = int(details['exit_status'])
        if td['exit_status'] != 0:
            td['status'] = fsl_sub.consts.FAILED
        td['error_message'] = None
        td['end_time'] = _qacct_timestamp(details['end_time'])
        td['start_time'] = _qacct_timestamp(details['start_time'])
        td['sub_time'] = _qacct_timestamp(details['qsub_time'])
        td['utime'] = float(details['ru_utime'].strip('s'))
        td['stime'] = float(details['ru_stime'].strip('s'))
        td['maxmemory'] = fsl_sub.utils.human_to_ram(
            details['maxvmem'], output='M')
    return job_details


def _qacct_timestamp(output):
    return datetime.datetime.strptime(output, '%a %b %d %H:%M:%S %Y')


def _qacct_job_lines(job_id):
    qacct = qacct_cmd()

    try:
        job_info = sp.run(
            [qacct, '-j', str(job_id)],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            check=True, universal_newlines=True)
        return job_info.stdout.split('\n')
    except FileNotFoundError:
        raise BadSubmission(
            "Grid Engine software may not be correctly installed")
    except sp.CalledProcessError as e:
        if "job id {} not found".format(job_id) in e.stderr:
            raise UnknownJobId(job_id)
        else:
            raise GridOutputError(e.stderr)


def _qstat_job(job_id, sub_job_id=None):
    job_details = {
        'exit_status': None,
        'error_message': None,
        'end_time': None, }
    task_id = 1
    sub_task = False
    if sub_job_id is not None:
        sub_task = True
        task_id = sub_job_id
    try:
        xml = ET.fromstring(_qstat_job_xml(job_id))
        djob_info = xml.find('djob_info')
        if djob_info is None:
            return None
        jobd = djob_info.find('element')
        jF = jobd.find
        job_details['id'] = int(jF('JB_job_number').text)
        job_details[
            'submission_time'] = datetime.datetime.fromtimestamp(
                int(jF('JB_submission_time').text))
        job_details['name'] = jF('JB_job_name').text
        job_details['job_directory'] = jF('JB_cwd').text
        predecessors = jF('JB_jid_predecessor_list')
        if predecessors:
            job_details['parents'] = [
                int(a.find('JRE_job_number').text) for a in
                predecessors.iter('job_predecessors')]
        else:
            job_details['parents'] = None
        offspring = jF('JB_jid_successor_list')
        if offspring:
            job_details['children'] = [
                int(a.find('JRE_job_number').text) for a in
                offspring.iter('ulong_sublist')]
        else:
            job_details['children'] = None
        job_details['script'] = jF('JB_script_file').text
        job_args = jF('JB_job_args')
        if job_args:
            job_details['args'] = [
                a.find('ST_name').text for a in
                jF('JB_job_args').iter('element')
            ]
        else:
            job_args = None
        job_tasks = jF('JB_ja_tasks')
        job_details['tasks'] = {}
        if job_tasks:
            # job is running or in Eqw
            sub_tasks = {}
            for task in job_tasks.iter('ulong_sublist'):
                # Problem is that JB_ja_tasks contains a nested ulong_sublist!
                task_number_object = task.find('JAT_task_number')
                if task_number_object is not None:
                    # We are in a top-level JB_ja_tasks element
                    task_number = int(task_number_object.text)

                    sc_u_l = task.find(
                        'JAT_scaled_usage_list'
                    )
                    if sc_u_l:
                        task_entry = {
                                'cpu': _get_scaled_value(sc_u_l, 'cpu'),
                                'maxvmem': _get_scaled_value(
                                    sc_u_l, 'maxvmem'),
                                'start_time': datetime.datetime.fromtimestamp(
                                    int(task.find('JAT_start_time').text)),

                            }
                    else:
                        task_entry = {
                            'cpu': None,
                            'maxvmem': None,
                            'starttime': None
                        }
                    err_mess = task.find('JAT_message_list')
                    task_entry['error_messages'] = []
                    if err_mess:
                        task_entry['status'] = (
                            fsl_sub.consts.FAILEDNQUEUED)
                        for message in err_mess.iter('ulong_sublist'):
                            task_entry['error_messages'].append(
                                        message.find(
                                            'QIM_message'
                                        ).text)
                    else:
                        task_entry['status'] = (
                            fsl_sub.consts.RUNNING)
                    sub_tasks[task_number] = task_entry
            if sub_task:
                job_details['tasks'] = {'1': sub_tasks[task_id], }
            else:
                job_details['tasks'] = sub_tasks
        else:
            # Job is pending
            job_details['tasks'][1] = {
                'cpu': None,
                'maxvmem': None,
                'start_time': None, }

            if predecessors:
                job_details['tasks'][1]['status'] = fsl_sub.consts.HELD
            else:
                job_details['tasks'][1]['status'] = fsl_sub.consts.QUEUED

    except Exception:
        raise GridOutputError("Unable to understand XML output")

    return job_details


def _get_scaled_value(xml_obj, name):
    for a in xml_obj.iter('scaled'):
        if a.find('UA_name').text == name:
            return a.find('UA_value').text


def _qstat_job_xml(job_id):
    qstat = qstat_cmd()
    output = None
    try:
        qstat_xml = sp.run(
            [qstat, '-j', str(job_id), '-xml'],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            check=True, universal_newlines=True)
        output = qstat_xml.stdout
    except FileNotFoundError:
        raise BadSubmission(
            "Grid Engine software may not be correctly installed")
    except sp.CalledProcessError as e:
        if "Following jobs do not exist:" in e.stderr:
            pass
        else:
            raise GridOutputError(e.stderr)
    return output


def project_list():
    '''This returns a list of recognised projects (or accounts) that a job
    can be allocated to (e.g. for billing or fair share allocation)'''
    qconf = qconf_cmd()
    try:
        qconf_out = sp.run(
            [qconf, '-sprjl'],
            stdout=sp.PIPE,
            stderr=sp.PIPE,
            check=True, universal_newlines=True
        )
    except FileNotFoundError:
        raise BadSubmission(
            "Grid Engine software may not be correctly installed")
    except sp.CalledProcessError as e:
        raise GridOutputError(e.stderr)
    return qconf_out.stdout.split()
